# ################################################################# #
#### LOAD LIBRARY AND DEFINE CORE SETTINGS                       ####
# ################################################################# #

### Initialise
rm(list = ls())
library(apollo)
apollo_initialise()

### Set core controls
apollo_control = list(
  modelName  = "timeUse",
  modelDescr = "Extended MDCEV model on time use.
                Leeds time use database",
  indivID    = "indivID",
  nCores     = 1
)

# ################################################################# #
#### LOAD DATA AND APPLY ANY TRANSFORMATIONS                     ####
# ################################################################# #

database = read.csv("apollo_timeUseData.csv")

### Create consumption variables for combined activities
# outside good: time spent at home and travelling
database$t_outside = rowSums(database[,c("t_a01", "t_a06", "t_a10", 
                                         "t_a11", "t_a12")]) 
database$t_leisure = rowSums(database[,c("t_a07", "t_a08", "t_a09")])


# ################################################################# #
#### DEFINE MODEL PARAMETERS                                     ####
# ################################################################# #

### Vector of parameters, including any that are kept fixed in estimation
y <- unlist(database[,paste0("t_",c("a02","a03","a04","a05","leisure"))]/60)
p <- 0.95
Q <- quantile(y[y!=0], p)
apollo_beta = c(aFemale      = 0, aWeekend     = 0, 
                gWork        = 1, gSchool      = 1, 
                gShopping    = 1, gPrivate     = 1, 
                gLeisure     = 1, bWork        = 0, 
                bSchool      = 0, bShopping    = 0, 
                bPrivate     = 0, bLeisure     = 0, 
                bWork_FT     = 0, bWork_wknd   = 0, 
                bSchool_young= 0, bLeisure_wknd= 0, 
                dWorkScho    = 0, dWorkShop    = 0, 
                dWorkPriv    = 0, dWorkLeis    = 0, 
                dSchoShop    = 0, dSchoPriv    = 0, 
                dSchoLeis    = 0, dShopPriv    = 0, 
                dShopLeis    = 0, dPrivLeis    = 0, 
                sig          = 1, 
                d0 = as.numeric(round(-log(1 - sqrt(p))/Q, 4)))
rm(y,p,Q)

### Name fixed parameters. c() for none
apollo_fixed = c("d0")

# ################################################################# #
#### GROUP AND VALIDATE INPUTS                                   ####
# ################################################################# #

apollo_inputs = apollo_validateInputs()
source("apollo_emdcev.R")

# ################################################################# #
#### DEFINE MODEL AND LIKELIHOOD FUNCTION                        ####
# ################################################################# #

apollo_probabilities=function(apollo_beta, apollo_inputs, 
                              functionality="estimate"){
    
  ### Initialise
  apollo_attach(apollo_beta, apollo_inputs)
  on.exit(apollo_detach(apollo_beta, apollo_inputs))
  P = list()
  
  ### Define individual alternatives
  alternatives  = c("work", "school", "shopping", "private", "leisure")
  
  ### List of ones
  nAlt = length(alternatives)
  ones = setNames(as.list(rep(1, nAlt)), alternatives)
  
  ### Define continuous consumption for individual alternatives
  X = list(work     =     t_a02/60,
           school   =     t_a03/60,
           shopping =     t_a04/60,
           private  =     t_a05/60,
           leisure  = t_leisure/60)
  
  ### Define utilities for individual alternatives
  V0= aFemale*female + aWeekend*weekend
  V = list()
  V[["work"]]     = bWork     + bWork_FT*occ_full_time + bWork_wknd*weekend
  V[["school"]]   = bSchool   + bSchool_young*(age<=30)
  V[["shopping"]] = bShopping
  V[["private"]]  = bPrivate
  V[["leisure"]]  = bLeisure  + bLeisure_wknd*weekend
  
  ### Define gamma parameters
  gamma = list(work     = gWork,    
               school   = gSchool,
               shopping = gShopping,
               private  = gPrivate,
               leisure  = gLeisure)
  
  ### Complementarity/substitution matrix
  delta <- c(0,                 0,         0,         0, 0,
             dWorkScho,         0,         0,         0, 0,
             dWorkShop, dSchoShop,         0,         0, 0,
             dWorkPriv, dSchoPriv, dShopPriv,         0, 0,
             dWorkLeis, dSchoLeis, dShopLeis, dPrivLeis, 0)
  delta <- matrix(delta, nrow=nAlt, ncol=nAlt, byrow=TRUE)
  delta0 <- list(work     = d0,    
                 school   = d0,
                 shopping = d0,
                 private  = d0,
                 leisure  = d0)
  
  ### Compute probabilities using eMDCEV model
  emdcev_settings <- list(X = X,
                          A = ones, 
                          V0= V0,
                          V = V, 
                          gamma = gamma, 
                          sigma = sig, 
                          delta0= delta0,
                          delta = delta, 
                          cost  = ones)
  P[["model"]] = apollo_emdcev(emdcev_settings, functionality)
  
  ### Process and return probabilities
  P = apollo_panelProd(P, apollo_inputs, functionality)
  P = apollo_prepareProb(P, apollo_inputs, functionality)
  return(P)
}

# ################################################################# #
#### MODEL ESTIMATION AND OUTPUT                                 ####
# ################################################################# #
s <- list(writeIter=FALSE)
model = apollo_estimate(apollo_beta, apollo_fixed, 
                        apollo_probabilities, apollo_inputs,
                        estimate_settings=s)

apollo_modelOutput(model)

s = list(saveEst=FALSE, saveCov=FALSE, saveCorr=FALSE)
apollo_saveOutput(model, saveOutput_settings=s)

# ################################################################# #
#### FORECAST AND FIT                                            ####
# ################################################################# #
model <- apollo_loadModel(apollo_control$modelName)
# Forecast
pred <- apollo_prediction(model, apollo_probabilities, apollo_inputs)

# Observation level level
x <- as.vector(pred[,3:7])
y <- database[,paste0("t_",c("a02","a03","a04","a05","leisure"))]/60
y <- unlist(setNames(y,NULL))
sqrt(mean((x-y)^2)) # 2.65

# Observation level per alternative
x <- pred[,3:7]
y <- as.matrix(database[,paste0("t_",c("a02","a03","a04","a05","leisure"))]/60)
round(sqrt(colMeans((x-y)^2)),2) # 7.44, 0.81, 2.00, 1.90, 4.82

# Sample level
x <- colSums(pred[,3:7])
y <- database[,paste0("t_",c("a02","a03","a04","a05","leisure"))]/60
y <- colSums(as.matrix(y))
sqrt(mean((x-y)^2)) # 959.57

# Outside / budget
x <- database$t_outside
x <- x/rowSums(database[,paste0("t_",c("a02","a03","a04","a05","leisure","outside"))])
summary(x) # 0.759
