# ############################################################# #
#### LOAD LIBRARY AND DEFINE CORE SETTINGS                   ####
# ############################################################# #

### Initialise
rm(list = ls())
library(apollo)
apollo_initialise()

### Set core controls
apollo_control = list(
  modelName  = "emdcevTrips",
  modelDescr = "Extended MDCEV model on number of trips.
                Chilean origin-destination survey 2012",
  indivID    = "hhId",
  nCores     = 5
)

# ############################################################# #
#### LOAD DATA AND APPLY ANY TRANSFORMATIONS                 ####
# ############################################################# #

database <- read.csv("apollo_odStgoTrips.csv")
# bicycle ownership is turned into a dichotomous variable
database$bicycle <- ifelse(database$nBicAdult>0, 1, 0)
# income in million of CLP
database$hhIncome <- database$hhIncome/10^6

# ############################################################# #
#### DEFINE MODEL PARAMETERS AND VALIDATE INPUT              ####
# ############################################################# #

### Vector of parameters, including any kept fixed in estimation
y = unlist(database[,c("work","stdy","perB",
                       "shop","leis","back")])
p = 0.95
Q = quantile(y[y!=0], p)
apollo_beta = c(bWo    =-0.0044, bWoHHSz= 0.0016, bWoNVeh= 0.0017, 
                bWoBicy= 0.0019, bWoHhIn= 0.0000, bWoNWor= 0.0165, 
                bSt    =-0.0254, bStHHSz= 0.0110, bStNVeh= 0.0035, 
                bStBicy= 0.0033, bStHhIn= 0.0000, bStNWor=-0.0060, 
                bPe    =-0.0028, bPeHHSz= 0.0043, bPeNVeh= 0.0052, 
                bPeBicy= 0.0000, bPeHhIn= 0.0010, bPeNWor=-0.0038, 
                bSh    =-0.0084, bShHHSz= 0.0040, bShNVeh= 0.0024, 
                bShBicy= 0.0014, bShHhIn= 0.0000, bShNWor=-0.0026, 
                bLe    =-0.0292, bLeHHSz= 0.0026, bLeNVeh= 0.0072, 
                bLeBicy= 0.0051, bLeHhIn= 0.0018, bLeNWor= 0.0000, 
                bBa    = 0.0035, bBaHHSz= 0.0076, bBaNVeh= 0.0027, 
                bBaBicy= 0.0015, bBaHhIn= 0.0000, bBaNWor= 0.0020, 
                gWork  =41.2179, gStdy  =51.9430, gPerB  =63.9743, 
                gShop  =51.2864, gLeis  =48.3174, gBack  =74.2110, 
                dWoSt  =-0.0142, dWoPe  =-0.0292, dStPe  =-0.0168, 
                dWoSh  =-0.0337, dStSh  =-0.0243, dPeSh  =-0.0219, 
                dWoLe  =-0.0110, dStLe  =-0.0139, dPeLe  =-0.0047, 
                dShLe  = 0.0000, dWoBa  = 0.0090, dStBa  = 0.0066, 
                dPeBa  = 0.0076, dShBa  = 0.0147, dLeBa  =-0.0234, 
                sig    = 0.0162, 
                d0 = as.numeric(round(-log(1 - sqrt(p))/Q, 4)))
rm(y,p,Q)
### Name fixed parameters. c() for none
apollo_fixed = c("d0", "bWoHhIn", "bStHhIn", "bPeBicy", 
                 "bShHhIn", "bLeNWor", "bBaHhIn", "dShLe")

# ############################################################# #
#### VALIDATE AND PREPARE INPUT                              ####
# ############################################################# #

apollo_inputs = apollo_validateInputs()
source("apollo_emdcev.R")

# ############################################################# #
#### DEFINE MODEL AND LIKELIHOOD FUNCTION                    ####
# ############################################################# #

apollo_probabilities=function(apollo_beta, apollo_inputs, 
                              functionality="estimate"){
  
  ### Initialise
  apollo_attach(apollo_beta, apollo_inputs)
  on.exit(apollo_detach(apollo_beta, apollo_inputs))
  P = list()
  
  ### MDCEV
  alts = c("work", "stdy", "perB", "shop", "leis", "back")
  nAlt = length(alts)
  X    = list(work=work, stdy=stdy, perB=perB,
              shop=shop, leis=leis, back=back)
  V0 = 0
  V  = list(
    work = bWo + bWoHHSz*hhSize + bWoNVeh*nVeh + bWoBicy*bicycle + bWoHhIn*hhIncome + bWoNWor*nWorkers,
    stdy = bSt + bStHHSz*hhSize + bStNVeh*nVeh + bStBicy*bicycle + bStHhIn*hhIncome + bStNWor*nWorkers,
    perB = bPe + bPeHHSz*hhSize + bPeNVeh*nVeh + bPeBicy*bicycle + bPeHhIn*hhIncome + bPeNWor*nWorkers,
    shop = bSh + bShHHSz*hhSize + bShNVeh*nVeh + bShBicy*bicycle + bShHhIn*hhIncome + bShNWor*nWorkers,
    leis = bLe + bLeHHSz*hhSize + bLeNVeh*nVeh + bLeBicy*bicycle + bLeHhIn*hhIncome + bLeNWor*nWorkers,
    back = bBa + bBaHHSz*hhSize + bBaNVeh*nVeh + bBaBicy*bicycle + bBaHhIn*hhIncome + bBaNWor*nWorkers)
  G    = list(work=gWork, stdy=gStdy, perB=gPerB,
              shop=gShop, leis=gLeis, back=gBack)
  ones = list(work=1, stdy=1, perB=1,
              shop=1, leis=1, back=1)
  delta  <- c(    0,     0,     0,     0,     0,     0,
              dWoSt,     0,     0,     0,     0,     0, 
              dWoPe, dStPe,     0,     0,     0,     0, 
              dWoSh, dStSh, dPeSh,     0,     0,     0, 
              dWoLe, dStLe, dPeLe, dShLe,     0,     0, 
              dWoBa, dStBa, dPeBa, dShBa, dLeBa,     0)
  delta  <- matrix(delta, nrow=nAlt, ncol=nAlt, byrow=TRUE)
  delta0 <- list(work=d0, stdy=d0, perB=d0,
                 shop=d0, leis=d0, back=d0)
  
  ### Compute probabilities using eMDCEV model
  emdcev_settings <- list(X = X,
                          A = ones, 
                          V0= V0,
                          V = V, 
                          gamma = G, 
                          sigma = sig, 
                          delta0= delta0,
                          delta = delta, 
                          cost  = ones)
  P[["model"]] = apollo_emdcev(emdcev_settings, functionality)
  
  ### Process and return probabilities
  P = apollo_prepareProb(P, apollo_inputs, functionality)
  return(P)
}

# ############################################################# #
#### MODEL ESTIMATION AND OUTPUT                             ####
# ############################################################# #

# Search for good starting values
#K <- length(apollo_beta); nb <- names(apollo_beta)
#minb <- setNames(rep(-1,K), nb)
#maxb <- setNames(rep( 1,K), nb)
#minb[grep("^g", nb)] <- 1
#maxb[grep("^g", nb)] <- 10
#minb[apollo_fixed] <- apollo_beta[apollo_fixed]
#maxb[apollo_fixed] <- apollo_beta[apollo_fixed]
#s <- list(nCandidates=5000,
#          apolloBetaMin=minb,
#          apolloBetaMax=maxb)
#apollo_beta <- apollo_searchStart(apollo_beta, apollo_fixed, 
#                                  apollo_probabilities,
#                                  apollo_inputs, s)
#rm(K, nb)

# Estimate
s <- list(maxIterations=500)
model = apollo_estimate(apollo_beta, apollo_fixed, 
                        apollo_probabilities, 
                        apollo_inputs, estimate_settings=s)

# Print and save results
apollo_modelOutput(model)
s = list(saveEst=FALSE, saveCov=FALSE, saveCorr=FALSE)
apollo_saveOutput(model, saveOutput_settings=s)

# ############################################################# #
#### FORECAST AND FIT                                        ####
# ############################################################# #

# Load model and data
model <- apollo_loadModel(apollo_control$modelName)

# Forecast
alts <- c("work", "stdy", "perB", "shop", "leis", "back")
pred <- apollo_prediction(model, apollo_probabilities, 
                          apollo_inputs)
x <- pred[,paste0(alts, "_ECont")]
y <- as.matrix(database[,alts])
# RMSE Observation level
sqrt(mean((x-y)^2)) # 
# RMSE Sample level
sqrt(mean((colSums(x)-colSums(y))^2)) # 
